%--------------------------------------------------------------------------
% Description: This function performs Principal Component Analysis for
% dimensionality reduction. It then whitens the retained principal
% components.
%
% Input:  data is a dxN matrix, where d is the dimensionality and N is the
%            number of samples.
%         max_principal_components is an integer indicating how many
%            principal components to retain.
%         skip_whitening defaults to false, but if it is set to true then
%            standard PCA is performed without whitening.
%
% Output: PCA_Mean is the mean of the data.
%         PCA_Vectors are the principal components.
%         PCA_Data is data after it has been projected onto the principal
%            components.
%
% To reduce the dimensionality of new data in a dxM matrix new_data, you
% would do the following: 
%   new_data_pca = PCA_Vectors*bsxfun(@minus, new_data, PCA_Mean);
%
% Example: d = 128; N = 1000;
%          data = randn(d, N);
%          new_data = randn(d, N);
%          [PCA_Mean, PCA_Vectors, PCA_Data] = Whitened_PCA(data, 3);
%          new_data_pca = PCA_Vectors*bsxfun(@minus, new_data, PCA_Mean);
%
% Author: Christopher Kanan (chriskanan@gmail.com)
%
%--------------------------------------------------------------------------
function [PCA_Mean, PCA_Vectors, PCA_Data] = Whitened_PCA(data, max_principal_components, skip_whitening)

if ~exist('skip_whitening', 'var')
    skip_whitening = false;
end

data(~isfinite(data)) = 0;

PCA_Mean = mean(data, 2);

max_principal_components = min(max_principal_components, size(data, 1));

%subtract mean
data = bsxfun(@minus, data, PCA_Mean);

data = data';
[U, sigma, PCA_Vectors] = svd(data, 0);

%sigma = diag(sigma);
data = data';

clear U
PCA_Vectors = PCA_Vectors';
max_principal_components = min(max_principal_components, size(PCA_Vectors, 1));
PCA_Vectors = PCA_Vectors(1:max_principal_components, :);

PCA_Vectors(~isreal(PCA_Vectors)) = 0;

%Whiten
if ~skip_whitening
    pca_data = PCA_Vectors * data;
    Whitening_Mat = (pca_data * pca_data');
    Whitening_Mat(Whitening_Mat < 0) = 0;
    Whitening_Mat = Whitening_Mat^(-1/2);
    Whitening_Mat(~isreal(Whitening_Mat)) = 0;
    PCA_Vectors = Whitening_Mat * PCA_Vectors;
    clear pca_data
    clear Whitening_Mat
end

PCA_Data = PCA_Vectors * data;        

return;
